@@ -88,7 +88,7 @@ pango_language_get_type (void)
 {
   static GType our_type = 0;
 
-  if (our_type == 0)
+  if (G_UNLIKELY (our_type == 0))
     our_type = g_boxed_type_register_static (I_("PangoLanguage"),
 					     (GBoxedCopyFunc)pango_language_copy,
 					     (GBoxedFreeFunc)pango_language_free);
@@ -434,7 +434,7 @@ script_for_lang_compare (gconstpointer key,
  * to write @language or if nothing is known about @language
  * (including the case that @language is %NULL),
  * %FALSE otherwise.
- *
+ 
  * Since: 1.4
  **/
 gboolean
@@ -715,7 +715,9 @@ pango_script_get_sample_language (PangoScript script)
   PangoLanguage *result;
 
   g_return_val_if_fail (script >= 0, NULL);
-  g_return_val_if_fail ((guint)script < G_N_ELEMENTS (sample_languages), NULL);
+
+  if ((guint)script >= G_N_ELEMENTS (sample_languages))
+    return NULL;
 
   result = _pango_script_get_default_language (script);
   if (result)
